import datetime
import json
import logging
import lzma
import os
import struct

import matplotlib.pyplot
import numpy
from PySide2 import QtCore, QtWidgets

import asphodel

from . import util
from ..info_dialog import InfoDialog
from ..setting_viewer import SettingViewerDialog

logger = logging.getLogger(__name__)


def get_save_file(default_name, parent=None):
    settings = QtCore.QSettings()

    # find the directory from settings
    directory = settings.value("fileSaveDirectory")
    if directory and type(directory) == str:
        if not os.path.isdir(directory):
            directory = None

    if not directory:
        directory = ""

    file_and_dir = os.path.join(directory, default_name)

    # ask the user for the file name
    caption = "Save File"
    file_filter = "Comma Seperated Value Files (*.csv);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(parent, caption, file_and_dir,
                                                file_filter)
    output_path = val[0]

    if output_path:
        # save the directory
        output_dir = os.path.dirname(output_path)
        settings.setValue("fileSaveDirectory", output_dir)
        return output_path
    else:
        return None


def file_information(parent=None):
    ret = util.load_single_file(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    _file_infos, header = ret

    dialog = InfoDialog(header)
    dialog.exec_()


def view_settings(parent=None):
    ret = util.load_single_file(parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    _file_infos, header = ret

    dialog = SettingViewerDialog(
        header['settings'], header['nvm'], header['custom_enums'],
        header['setting_categories'], parent=parent)
    ret = dialog.exec_()


def raw_export(parent=None):
    filename = util.get_packdata_file(parent)
    if filename is None:
        return  # user cancelled

    try:
        # open the compressed file, and figure out how many bytes it has
        rawfile = open(filename, 'rb')
        rawfile.seek(0, os.SEEK_END)
        rawlen = rawfile.tell()
        rawfile.seek(0, os.SEEK_SET)

        # open the LZMA wrapper object
        lzmafile = lzma.LZMAFile(rawfile, "rb")

        # read the header
        header_leader = struct.unpack(">dI", lzmafile.read(12))
        header_bytes = lzmafile.read(header_leader[1])

        if len(header_bytes) == 0:
            message = "Empty header in {}!".format(filename)
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return  # error

        try:
            header_str = header_bytes.decode("UTF-8")
            header = json.loads(header_str)
        except:
            message = "Could not parse file header!"
            logger.exception(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return
    except:
        message = "Could not open file {}!".format(filename)
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return  # error

    # ask the user for the json file name
    default_json_filename = os.path.splitext(filename)[0] + ".json"
    caption = "Save JSON File"
    json_file_filter = "JSON Files (*.json);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(
        parent, caption, default_json_filename, json_file_filter)
    json_filename = val[0]

    if json_filename:
        with open(json_filename, 'wt') as f:
            json.dump(header, f, sort_keys=True, indent=4)

    # ask the user for the packet file name
    default_packet_filename = os.path.splitext(filename)[0] + ".packet"
    caption = "Save Packet File"
    packet_file_filter = "Packet Files (*.packet);;All Files (*.*)"
    val = QtWidgets.QFileDialog.getSaveFileName(
        parent, caption, default_packet_filename, packet_file_filter)
    packet_filename = val[0]

    if not packet_filename:
        # user didn't want to save a packet file, just exit now
        return

    packet_file = open(packet_filename, 'wt')

    packet_leader = struct.Struct(">dI")
    stream_packet_length = header['stream_packet_length']

    # create a progress dialog box
    window_flags = (QtCore.Qt.MSWindowsFixedSizeDialogHint)
    progress = QtWidgets.QProgressDialog("Loading Data...", "Abort", 0,
                                         rawlen, parent, window_flags)
    progress.setWindowModality(QtCore.Qt.WindowModal)
    progress.forceShow()

    # make sure it gets at least one value different than rawlen
    progress.setValue(0)

    process_countdown = 0

    try:
        while True:
            leader_bytes = lzmafile.read(packet_leader.size)

            if not leader_bytes:
                break  # file is finished

            leader = packet_leader.unpack(leader_bytes)
            packet_timestamp = leader[0]
            packet_bytes = lzmafile.read(leader[1])

            packet_dt = datetime.datetime.fromtimestamp(packet_timestamp,
                                                        datetime.timezone.utc)
            packet_time_string = "{} ".format(
                packet_dt.strftime('%Y-%m-%dT%H:%M:%S.%f'))

            for i in range(len(packet_bytes) // stream_packet_length):
                start = i * stream_packet_length
                packet = packet_bytes[start:start + stream_packet_length]
                packet_string = ", ".join(map("0x{:02x}".format, packet))

                packet_file.write(packet_time_string)
                packet_file.write('[' + packet_string + ']\n')

            progress.setValue(rawfile.tell())

            # update the window periodically
            if process_countdown == 0:
                process_countdown = 100
                QtCore.QCoreApplication.processEvents()
            else:
                process_countdown -= 1

            # abort if required
            if progress.wasCanceled():
                return  # user cancelled
    except:
        progress.cancel()
        m = "Error while processing data in file {}!".format(filename)
        logger.exception(m)
        QtWidgets.QMessageBox.critical(parent, "Error", m)
        return

    logger.debug("Finished reading {}.".format(filename))


def split_file(parent=None):
    sequences = []
    unit_types = []
    unit_names = []

    # load the file
    ret = util.load_single_file(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    subchannels_list = util.choose_subchannels(header, parent=parent)
    if subchannels_list is None:
        return  # error message displayed, or user cancelled

    # get all valid channel indexes
    channel_indexes = sorted({t[0] for t in subchannels_list})

    batch_info = util.decode_batch(file_infos, header, channel_indexes,
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    sequence_info, unit_info = util.sequence_data(batch_info, chunk=False,
                                                  parent=parent)

    for channel_index in channel_indexes:
        sequences.append(sequence_info[channel_index][0])
        unit_types.append(header['channels'][channel_index].unit_type)
        unit_names.append(unit_info[channel_index]['utf8'])

    # find the earliest start time
    min_start = min(s[2] for s in sequences)

    # group channels by unit type
    plot_groups = []
    groups_by_unit_type = {}
    for batch_index, unit_type in enumerate(unit_types):
        if unit_type == 0:
            # don't share plots between channels with unknown units
            plot_groups.append([batch_index])
        elif unit_type in groups_by_unit_type:
            # already have a group made
            plot_group = groups_by_unit_type[unit_type]
            plot_group.append(batch_index)
        else:
            # need a new group
            plot_group = [batch_index]
            plot_groups.append(plot_group)
            groups_by_unit_type[unit_type] = plot_group

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax_list = matplotlib.pyplot.subplots(len(plot_groups), squeeze=False,
                                              sharex=True)

    for i, ax in enumerate(ax_list[:, 0]):
        plot_group = plot_groups[i]

        legend_strs = []

        for batch_index in plot_group:
            time, data, start, _end = sequences[batch_index]

            delta = start - min_start

            ax.plot(time + delta, data)

            channel_index = channel_indexes[batch_index]
            channel_data = batch_info[channel_index]
            legend_strs.extend(channel_data.channel_decoder.subchannel_names)

        ax.legend()

        ax.set_xlabel("Time (s)")
        unit_str = unit_names[plot_group[0]]
        if unit_str:
            ax.set_ylabel("Magnitude ({})".format(unit_str))
        else:
            ax.set_ylabel("Magnitude")

        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)

        ax.legend(legend_strs)

    # have the user select points
    ax_list[0, 0].set_title("Select split points. Done = right click / enter\n"
                            "Undo = middle click / backspace")
    fig.show()
    points = fig.ginput(n=-1, timeout=0, mouse_add=1, mouse_stop=3,
                        mouse_pop=2)
    matplotlib.pyplot.close(fig)

    if len(points) == 0:
        return  # cancelled

    split_times = [p[0] for p in points]

    logger.debug("Selected split times: %s", split_times)

    split_timestamps = []

    for split_time in split_times:
        min_ts = None
        for channel_index, channel_data in sorted(batch_info.items()):
            delta = channel_data.timestamps[0] - min_start

            target_index = (split_time - delta) * channel_data.stream.rate
            i = numpy.searchsorted(channel_data.indexes, target_index)
            timestamps = channel_data.timestamps[i - 1:i + 1]
            if len(timestamps) == 2:
                ts = min(timestamps)
                if min_ts is None or ts < min_ts:
                    min_ts = ts
        split_timestamps.append(min_ts)

    bad_element_count = sum(ts is None for ts in split_timestamps)
    if bad_element_count == len(split_timestamps):
        message = "No valid split points"
        logger.error(message)
        QtWidgets.QMessageBox.critical(parent, "Invalid Points", message)
        return
    elif bad_element_count > 0:
        message = "Selected {} invalid split point{}".format(
            bad_element_count, "s" if bad_element_count != 1 else "")
        logger.info(message)
        QtWidgets.QMessageBox.information(parent, "Invalid Points", message)

    split_timestamps = sorted(ts for ts in split_timestamps if ts is not None)
    logger.debug("Selected split timestamps: %s", split_timestamps)

    output_file_count = len(split_timestamps) + 1

    # do the actual splitting
    filename = file_infos[0][0]
    root, ext = os.path.splitext(filename)
    if ext == ".apd" or ext == ".apdbak":
        base_filename = root
    else:
        base_filename = filename

    try:
        # open the compressed file, and figure out how many bytes it has
        rawfile = open(filename, 'rb')
        rawfile.seek(0, os.SEEK_END)
        rawlen = rawfile.tell()
        rawfile.seek(0, os.SEEK_SET)

        # open the LZMA wrapper object
        lzmafile = lzma.LZMAFile(rawfile, "rb")

        # read the header
        header_leader = struct.unpack(">dI", lzmafile.read(12))
        header_bytes = lzmafile.read(header_leader[1])
    except:
        message = "Could not open file {}!".format(filename)
        logger.exception(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return  # error

    packet_leader = struct.Struct(">dI")

    # create a progress dialog box
    window_flags = (QtCore.Qt.MSWindowsFixedSizeDialogHint)
    progress = QtWidgets.QProgressDialog("Processing Data...", "Abort", 0,
                                         rawlen, parent, window_flags)
    progress.setWindowModality(QtCore.Qt.WindowModal)
    progress.forceShow()

    # make sure it gets at least one value different than rawlen
    progress.setValue(0)

    process_countdown = 0

    next_dest_index = 1
    dest_file = None

    try:
        while True:
            leader_bytes = lzmafile.read(packet_leader.size)

            if not leader_bytes:
                break  # file is finished

            leader = packet_leader.unpack(leader_bytes)
            packet_timestamp = leader[0]
            packet_bytes = lzmafile.read(leader[1])

            if split_timestamps and packet_timestamp > split_timestamps[0]:
                if dest_file is not None:
                    dest_file.close()
                dest_file = None
                del split_timestamps[0]

            if dest_file is None:
                # open a new file
                if dest_file is not None:
                    dest_file.close()
                dest_filename = "{}_{}.apd".format(base_filename,
                                                   next_dest_index)
                next_dest_index += 1
                dest_file = lzma.open(dest_filename, 'wb')

                dest_file.write(struct.pack(">dI", *header_leader))
                dest_file.write(header_bytes)

                logger.debug("opened file %s", dest_filename)

            # write the data into the file
            dest_file.write(leader_bytes)
            dest_file.write(packet_bytes)

            progress.setValue(rawfile.tell())

            # update the window periodically
            if process_countdown == 0:
                process_countdown = 100
                QtCore.QCoreApplication.processEvents()
            else:
                process_countdown -= 1

            # abort if required
            if progress.wasCanceled():
                return  # user cancelled
    except:
        progress.cancel()
        m = "Error while processing data in file {}!".format(filename)
        logger.exception(m)
        QtWidgets.QMessageBox.critical(parent, "Error", m)
        return

    progress.reset()
    if dest_file is not None:
        dest_file.close()

    lzmafile.close()
    rawfile.close()

    logger.debug("Finished processing {}.".format(filename))

    os.rename(filename, base_filename + ".apdbak")

    m = "Split into {} files".format(output_file_count)
    QtWidgets.QMessageBox.information(parent, "Finished", m)



# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os.path

from PySide2 import QtGui, QtWidgets

from . import mondo_rc  # needed for the icon
from .analysis.csv import get_csv_file


def _do_export(filename, labels, xdata, ydata):
    with open(filename, "w") as f:
        f.write(", ".join(labels))
        f.write("\n")

        for x, y in zip(xdata, ydata):
            f.write("{}, {}\n".format(x, y))


def add_export_csv_action(figure, labels, data):
    toolbar = figure.canvas.toolbar
    actionText = QtWidgets.QApplication.translate("ExportCSVAction",
                                                  "Export CSV")
    action = QtWidgets.QAction(actionText, toolbar)
    icon = QtGui.QIcon()
    icon.addPixmap(QtGui.QPixmap(":/icon/icons/export-csv.png"))
    action.setIcon(icon)

    def handle_export():
        base_file = get_csv_file("export", parent=toolbar)
        if not base_file:
            return

        root, ext = os.path.splitext(base_file)

        for xdata, ydata, name in data:
            if len(data) == 1:
                filename = base_file
            else:
                filename = root + "_" + name + ext

            _do_export(filename, labels, xdata, ydata)

    action.triggered.connect(handle_export)

    toolbar.addAction(action)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'info_dialog.ui',
# licensing of 'info_dialog.ui' applies.
#
# Created: Tue Apr  2 19:36:21 2019
#      by: pyside2-uic  running on PySide2 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_InfoDialog(object):
    def setupUi(self, InfoDialog):
        InfoDialog.setObjectName("InfoDialog")
        InfoDialog.resize(736, 548)
        self.verticalLayout = QtWidgets.QVBoxLayout(InfoDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.plainTextEdit = QtWidgets.QPlainTextEdit(InfoDialog)
        self.plainTextEdit.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.plainTextEdit.setReadOnly(True)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout.addWidget(self.plainTextEdit)
        self.buttonBox = QtWidgets.QDialogButtonBox(InfoDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(InfoDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), InfoDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), InfoDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(InfoDialog)

    def retranslateUi(self, InfoDialog):
        InfoDialog.setWindowTitle(QtWidgets.QApplication.translate("InfoDialog", "Device Information", None, -1))




# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import importlib
import logging
import os
import subprocess
import sys
import tempfile
import threading

from PySide2 import QtCore, QtWidgets

import asphodel
import hyperborea.download

from .ui_main import Ui_MondoMainWindow
from .about import AboutDialog
from .analysis import csv, file, psd, spectrogram, time
from .preferences import PreferencesDialog

logger = logging.getLogger(__name__)


class MondoMainWindow(QtWidgets.QMainWindow, Ui_MondoMainWindow):
    find_update_finished = QtCore.Signal(object)
    update_download_finished = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.update_progress_lock = threading.Lock()

        self.setupUi(self)

        self.extra_ui_setup()
        self.setup_callbacks()
        self.setup_update_actions()

    def extra_ui_setup(self):
        self.update_progress = QtWidgets.QProgressDialog("", "", 0, 100)
        self.update_progress.setLabelText(self.tr(""))
        self.update_progress.setWindowTitle(self.tr("Check for Update"))
        self.update_progress.setCancelButton(None)
        self.update_progress.setWindowModality(QtCore.Qt.WindowModal)
        self.update_progress.setMinimumDuration(0)
        self.update_progress.setAutoReset(False)
        self.update_progress.reset()

    def setup_callbacks(self):
        self.timeButton.clicked.connect(self.time_analysis)
        self.synchronousTimeButton.clicked.connect(
            self.synchronous_time_analysis)
        self.overlaidTimeButton.clicked.connect(self.overlaid_time_analysis)
        self.spectrogramButton.clicked.connect(self.spectrogram_analysis)
        self.psdButton.clicked.connect(self.psd_analysis)
        self.singleChannelPsdButton.clicked.connect(
            self.single_channel_psd_analysis)
        self.singleSlicePsdButton.clicked.connect(
            self.single_slice_psd_analysis)
        self.overlaidPsdButton.clicked.connect(self.overlaid_psd_analysis)
        self.csvButton.clicked.connect(self.csv_export)
        self.csvDownsampledButton.clicked.connect(self.csv_downsampled_export)
        self.fileInformationButton.clicked.connect(self.file_information)
        self.viewSettingsButton.clicked.connect(self.view_settings)
        self.rawExportButton.clicked.connect(self.raw_export)
        self.splitFileButton.clicked.connect(self.split_file)

        self.actionPreferences.triggered.connect(self.show_preferences)

        self.actionAbout.triggered.connect(self.show_about)
        self.actionAboutLibraries.triggered.connect(self.show_about_libraries)

        self.actionUpdateLatestStable.triggered.connect(
            self.update_latest_stable)
        self.actionUpdateCurrentBranch.triggered.connect(
            self.update_current_branch)
        self.actionUpdateSpecificBranch.triggered.connect(
            self.update_specific_branch)
        self.actionUpdateSpecificCommit.triggered.connect(
            self.update_specific_commit)

        self.update_progress_timer = QtCore.QTimer(self)
        self.update_progress_timer.timeout.connect(self.update_progress_cb)

        self.software_finder = hyperborea.download.SoftwareFinder()
        self.software_finder.completed.connect(self.update_finder_completed)
        self.software_finder.error.connect(self.update_finder_error)
        self.software_downloader = hyperborea.download.Downloader()
        self.software_downloader.completed.connect(
            self.software_download_completed)
        self.software_downloader.error.connect(self.software_download_error)

    def setup_update_actions(self):
        is_frozen = getattr(sys, 'frozen', False)

        valid_info = False
        if is_frozen:
            # load the build_info.txt
            main_dir = os.path.dirname(sys.executable)
            build_info_filename = os.path.join(main_dir, "build_info.txt")
            try:
                with open(build_info_filename, "r") as f:
                    lines = f.readlines()
                    self.branch_name = lines[0].strip()
                    self.commit_hash = lines[1].strip()
                    self.build_key = lines[2].strip()
                    valid_info = True
            except:
                logger.exception('Could not read build_info.txt')

        if not valid_info:
            self.menuCheckForUpdates.setEnabled(False)
            self.menuCheckForUpdates.setTitle(self.tr("Not Updatable"))
            self.actionUpdateLatestStable.setEnabled(False)
            self.actionUpdateCurrentBranch.setEnabled(False)
            self.actionUpdateSpecificBranch.setEnabled(False)
            self.actionUpdateSpecificCommit.setEnabled(False)
        else:
            if self.branch_name == "master":
                # master is latest stable
                self.actionUpdateCurrentBranch.setEnabled(False)
                self.actionUpdateCurrentBranch.setVisible(False)
            else:
                action_str = self.tr("Latest {}").format(self.branch_name)
                self.actionUpdateCurrentBranch.setText(action_str)

    def find_update(self, branch=None, commit=None):
        self.update_progress.setMinimum(0)
        self.update_progress.setMaximum(0)
        self.update_progress.setValue(0)
        self.update_progress.setLabelText(self.tr("Checking for update..."))
        self.update_progress.forceShow()

        self.software_finder.find_software("mondo", self.build_key, branch,
                                           commit)

    def update_finder_error(self, error_str):
        self.update_progress.reset()
        QtWidgets.QMessageBox.critical(self, self.tr("Error"), error_str)

    def update_finder_completed(self, params):
        self.update_progress.reset()
        url, commit, ready = params

        if commit is not None and commit == self.commit_hash:
            logger.info("Up to date with commit %s", commit)
            QtWidgets.QMessageBox.information(
                self, self.tr("Up to date"),
                self.tr("Already running this version"))
            return

        if not ready:
            logger.info("Update is not ready")
            QtWidgets.QMessageBox.information(
                self, self.tr("Update not ready"),
                self.tr("Update is not ready"))
            return

        # ask if the user wants to proceed
        ret = QtWidgets.QMessageBox.question(
            self, self.tr("Update?"), self.tr("Update available. Update now?"),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if ret != QtWidgets.QMessageBox.Yes:
            return

        logger.info("Downloading update from %s", url)

        self.update_progress.setMinimum(0)
        self.update_progress.setMaximum(0)
        self.update_progress.setValue(0)
        self.update_progress.setLabelText(self.tr("Downloading update..."))
        self.update_progress.forceShow()

        self.update_progress_written = 0
        self.update_progress_total = 0

        fd, filename = tempfile.mkstemp(".exe", "setup-", text=False)
        file = os.fdopen(fd, "wb")
        file.filename = filename
        self.software_downloader.start_download(url, file,
                                                self.update_progress_func)

        self.update_progress_timer.start(20)  # 20 milliseconds

    def update_progress_func(self, written_bytes, total_length):
        with self.update_progress_lock:
            self.update_progress_total = total_length
            self.update_progress_written = written_bytes

    def update_progress_cb(self):
        with self.update_progress_lock:
            if self.update_progress_total != 0:
                self.update_progress.setMinimum(0)
                self.update_progress.setMaximum(self.update_progress_total)
                self.update_progress.setValue(self.update_progress_written)

    def software_download_error(self, file, error_str):
        self.update_progress_timer.stop()
        self.update_progress.reset()
        file.close()
        QtWidgets.QMessageBox.critical(self, self.tr("Error"), error_str)
        os.unlink(file.filename)

    def software_download_completed(self, file):
        self.update_progress_timer.stop()
        self.update_progress.reset()
        file.close()

        # run the intstaller
        subprocess.Popen([file.filename, '/silent', "/DeleteInstaller=Yes",
                          "/SP-", "/SUPPRESSMSGBOXES", "/NORESTART",
                          "/NOCANCEL"])

        # close the application (though installer will force kill regardless)
        self.close()

    def update_latest_stable(self):
        self.find_update(branch="master")

    def update_current_branch(self):
        self.find_update(branch=self.branch_name)

    def update_specific_branch(self):
        branch, ok = QtWidgets.QInputDialog.getText(
            self, self.tr("Branch"), self.tr("Branch:"),
            QtWidgets.QLineEdit.Normal, "master")
        if not ok:
            return

        branch = branch.strip()

        self.find_update(branch=branch)

    def update_specific_commit(self):
        commit, ok = QtWidgets.QInputDialog.getText(
            self, self.tr("Commit"), self.tr("Commit:"),
            QtWidgets.QLineEdit.Normal, "")
        if not ok:
            return

        commit = commit.strip()

        self.find_update(commit=commit)

    def show_about(self):
        dialog = AboutDialog(self)
        dialog.exec_()

    def _get_version(self, library):
        try:
            lib = importlib.import_module(library)
            return lib.__version__
        except (AttributeError, ImportError):
            return "ERROR"

    def show_about_libraries(self):
        libraries = ["numpy", "scipy", "PySide2", "matplotlib", 'hyperborea']
        vers = {}
        for lib in libraries:
            vers[lib] = self._get_version(lib)

        # special case for asphodel and asphodel_py
        vers["asphodel_py"] = self._get_version("asphodel")
        vers["asphodel"] = asphodel.build_info

        # python version (sys.version is too long)
        is_64bit = sys.maxsize > (2 ** 32)
        bit_str = "64 bit" if is_64bit else "32 bit"
        python_ver = ".".join(map(str, sys.version_info[:3]))
        python_str = "{} ({} {})".format(python_ver, sys.platform, bit_str)
        vers['python'] = python_str

        s = "\n".join(k + ": " + vers[k] for k in sorted(vers, key=str.lower))
        QtWidgets.QMessageBox.information(self, "About Libraries", s)

    def do_analysis(self, analysis, *args, **kwargs):
        try:
            analysis(self, *args, **kwargs)
        except:
            logger.exception("Uncaught exception")
            QtWidgets.QMessageBox.critical(
                self, "Error", "Uncaught exception. See log for details.")

    def time_analysis(self):
        self.do_analysis(time.time_analysis)

    def synchronous_time_analysis(self):
        self.do_analysis(time.synchronous_time_analysis)

    def overlaid_time_analysis(self):
        self.do_analysis(time.overlaid_time_analysis)

    def spectrogram_analysis(self):
        self.do_analysis(spectrogram.spectrogram_analysis)

    def psd_analysis(self):
        self.do_analysis(psd.psd_analysis)

    def single_channel_psd_analysis(self):
        self.do_analysis(psd.single_channel_psd_analysis)

    def single_slice_psd_analysis(self):
        self.do_analysis(psd.single_slice_psd_analysis)

    def overlaid_psd_analysis(self):
        self.do_analysis(psd.overlaid_psd_analysis)

    def csv_export(self):
        self.do_analysis(csv.csv_export)

    def csv_downsampled_export(self):
        self.do_analysis(csv.csv_export, downsample=True)

    def file_information(self):
        self.do_analysis(file.file_information)

    def view_settings(self):
        self.do_analysis(file.view_settings)

    def raw_export(self):
        self.do_analysis(file.raw_export)

    def split_file(self):
        self.do_analysis(file.split_file)

    def show_preferences(self):
        dialog = PreferencesDialog(self)
        dialog.exec_()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'datetime_subset.ui',
# licensing of 'datetime_subset.ui' applies.
#
# Created: Mon Mar  2 17:09:12 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_DateTimeSubsetDialog(object):
    def setupUi(self, DateTimeSubsetDialog):
        DateTimeSubsetDialog.setObjectName("DateTimeSubsetDialog")
        DateTimeSubsetDialog.resize(283, 284)
        self.gridLayout = QtWidgets.QGridLayout(DateTimeSubsetDialog)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 4, 0, 1, 3)
        self.useAllData = QtWidgets.QRadioButton(DateTimeSubsetDialog)
        self.useAllData.setChecked(True)
        self.useAllData.setObjectName("useAllData")
        self.gridLayout.addWidget(self.useAllData, 0, 0, 1, 3)
        self.buttonBox = QtWidgets.QDialogButtonBox(DateTimeSubsetDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 5, 0, 1, 3)
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_2.setObjectName("formLayout_2")
        self.startLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.startLabel.setEnabled(False)
        self.startLabel.setObjectName("startLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.startLabel)
        self.startDateTime = QtWidgets.QDateTimeEdit(DateTimeSubsetDialog)
        self.startDateTime.setEnabled(False)
        self.startDateTime.setWrapping(True)
        self.startDateTime.setObjectName("startDateTime")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.startDateTime)
        self.endLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.endLabel.setEnabled(False)
        self.endLabel.setObjectName("endLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.endLabel)
        self.endDateTime = QtWidgets.QDateTimeEdit(DateTimeSubsetDialog)
        self.endDateTime.setEnabled(False)
        self.endDateTime.setWrapping(True)
        self.endDateTime.setObjectName("endDateTime")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.endDateTime)
        self.durationLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.durationLabel.setEnabled(False)
        self.durationLabel.setObjectName("durationLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.durationLabel)
        self.label = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.label.setEnabled(False)
        self.label.setObjectName("label")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label)
        self.label_2 = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.label_2.setEnabled(False)
        self.label_2.setObjectName("label_2")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.startSeconds = QtWidgets.QSpinBox(DateTimeSubsetDialog)
        self.startSeconds.setEnabled(False)
        self.startSeconds.setObjectName("startSeconds")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.startSeconds)
        self.endSeconds = QtWidgets.QSpinBox(DateTimeSubsetDialog)
        self.endSeconds.setEnabled(False)
        self.endSeconds.setObjectName("endSeconds")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.endSeconds)
        self.duration = QtWidgets.QSpinBox(DateTimeSubsetDialog)
        self.duration.setEnabled(False)
        self.duration.setObjectName("duration")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.duration)
        self.gridLayout.addLayout(self.formLayout_2, 3, 1, 1, 2)
        spacerItem1 = QtWidgets.QSpacerItem(13, 13, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 3, 0, 1, 1)
        self.useSubset = QtWidgets.QRadioButton(DateTimeSubsetDialog)
        self.useSubset.setObjectName("useSubset")
        self.gridLayout.addWidget(self.useSubset, 2, 0, 1, 3)
        self.formLayout_4 = QtWidgets.QFormLayout()
        self.formLayout_4.setObjectName("formLayout_4")
        self.allStartLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allStartLabel.setObjectName("allStartLabel")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.allStartLabel)
        self.allStart = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allStart.setObjectName("allStart")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.allStart)
        self.allEndLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allEndLabel.setObjectName("allEndLabel")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.allEndLabel)
        self.allEnd = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allEnd.setObjectName("allEnd")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.allEnd)
        self.allDurationLabel = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allDurationLabel.setObjectName("allDurationLabel")
        self.formLayout_4.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.allDurationLabel)
        self.allDuration = QtWidgets.QLabel(DateTimeSubsetDialog)
        self.allDuration.setObjectName("allDuration")
        self.formLayout_4.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.allDuration)
        self.gridLayout.addLayout(self.formLayout_4, 1, 1, 1, 2)

        self.retranslateUi(DateTimeSubsetDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), DateTimeSubsetDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), DateTimeSubsetDialog.reject)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.startLabel.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.startDateTime.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.endLabel.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.endDateTime.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.durationLabel.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allStartLabel.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allStart.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allEndLabel.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allEnd.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allDurationLabel.setEnabled)
        QtCore.QObject.connect(self.useAllData, QtCore.SIGNAL("toggled(bool)"), self.allDuration.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.label.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.label_2.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.startSeconds.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.endSeconds.setEnabled)
        QtCore.QObject.connect(self.useSubset, QtCore.SIGNAL("toggled(bool)"), self.duration.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(DateTimeSubsetDialog)
        DateTimeSubsetDialog.setTabOrder(self.useAllData, self.useSubset)
        DateTimeSubsetDialog.setTabOrder(self.useSubset, self.startDateTime)
        DateTimeSubsetDialog.setTabOrder(self.startDateTime, self.endDateTime)
        DateTimeSubsetDialog.setTabOrder(self.endDateTime, self.buttonBox)

    def retranslateUi(self, DateTimeSubsetDialog):
        DateTimeSubsetDialog.setWindowTitle(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Select Data Range", None, -1))
        self.useAllData.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Use all data", None, -1))
        self.startLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Start Time (UTC):", None, -1))
        self.startDateTime.setDisplayFormat(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "M/d/yyyy h:mm:ss AP", None, -1))
        self.endLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "End Time (UTC):", None, -1))
        self.endDateTime.setDisplayFormat(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "M/d/yyyy h:mm:ss AP", None, -1))
        self.durationLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Duration:", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Relative Start Time", None, -1))
        self.label_2.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Relative End Time", None, -1))
        self.startSeconds.setSuffix(QtWidgets.QApplication.translate("DateTimeSubsetDialog", " s", None, -1))
        self.endSeconds.setSuffix(QtWidgets.QApplication.translate("DateTimeSubsetDialog", " s", None, -1))
        self.duration.setSuffix(QtWidgets.QApplication.translate("DateTimeSubsetDialog", " s", None, -1))
        self.useSubset.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Use data within range", None, -1))
        self.allStartLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Start Time (UTC):", None, -1))
        self.allStart.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "1/1/2000 12:00:00 AM", None, -1))
        self.allEndLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "End Time (UTC):", None, -1))
        self.allEnd.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "1/1/2000 12:00:00 AM", None, -1))
        self.allDurationLabel.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "Duration:", None, -1))
        self.allDuration.setText(QtWidgets.QApplication.translate("DateTimeSubsetDialog", "0 s", None, -1))




from PySide2 import QtCore, QtWidgets

from .ui_preferences import Ui_PreferencesDialog


class PreferencesDialog(QtWidgets.QDialog, Ui_PreferencesDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.settings = QtCore.QSettings()

        self.setupUi(self)

        # this is easily forgotten in Qt Designer
        self.tabWidget.setCurrentIndex(0)

        self.accepted.connect(self.write_settings)

        self.read_settings()

    def read_settings(self):
        self.unitPreferences.read_settings()

    def write_settings(self):
        self.unitPreferences.write_settings()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'psd_options.ui',
# licensing of 'psd_options.ui' applies.
#
# Created: Tue Apr  2 19:36:21 2019
#      by: pyside2-uic  running on PySide2 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_PSDOptionsWidget(object):
    def setupUi(self, PSDOptionsWidget):
        PSDOptionsWidget.setObjectName("PSDOptionsWidget")
        PSDOptionsWidget.resize(208, 191)
        self.formLayout = QtWidgets.QFormLayout(PSDOptionsWidget)
        self.formLayout.setContentsMargins(0, 0, 0, 0)
        self.formLayout.setObjectName("formLayout")
        self.fftPointsLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.fftPointsLabel.setObjectName("fftPointsLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.fftPointsLabel)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.fftPoints = QtWidgets.QComboBox(PSDOptionsWidget)
        self.fftPoints.setObjectName("fftPoints")
        self.horizontalLayout_4.addWidget(self.fftPoints)
        self.formLayout.setLayout(0, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_4)
        self.durationLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.durationLabel.setObjectName("durationLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.durationLabel)
        self.duration = QtWidgets.QLabel(PSDOptionsWidget)
        self.duration.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.duration.setObjectName("duration")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.duration)
        self.resolutionLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.resolutionLabel.setObjectName("resolutionLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.resolutionLabel)
        self.resolution = QtWidgets.QLabel(PSDOptionsWidget)
        self.resolution.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.resolution.setObjectName("resolution")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.resolution)
        self.overlapPercentLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.overlapPercentLabel.setObjectName("overlapPercentLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.overlapPercentLabel)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.overlapPercent = QtWidgets.QDoubleSpinBox(PSDOptionsWidget)
        self.overlapPercent.setMinimumSize(QtCore.QSize(75, 0))
        self.overlapPercent.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.overlapPercent.setDecimals(0)
        self.overlapPercent.setMaximum(95.0)
        self.overlapPercent.setSingleStep(5.0)
        self.overlapPercent.setProperty("value", 75.0)
        self.overlapPercent.setObjectName("overlapPercent")
        self.horizontalLayout_3.addWidget(self.overlapPercent)
        self.formLayout.setLayout(3, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_3)
        self.overlapPointsLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.overlapPointsLabel.setObjectName("overlapPointsLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.overlapPointsLabel)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem2)
        self.overlapPoints = QtWidgets.QSpinBox(PSDOptionsWidget)
        self.overlapPoints.setMinimumSize(QtCore.QSize(75, 0))
        self.overlapPoints.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.overlapPoints.setObjectName("overlapPoints")
        self.horizontalLayout_5.addWidget(self.overlapPoints)
        self.formLayout.setLayout(4, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_5)
        self.windowCountLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.windowCountLabel.setObjectName("windowCountLabel")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.windowCountLabel)
        self.windowCount = QtWidgets.QLabel(PSDOptionsWidget)
        self.windowCount.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.windowCount.setObjectName("windowCount")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.windowCount)
        self.windowFunctionLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.windowFunctionLabel.setObjectName("windowFunctionLabel")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.windowFunctionLabel)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem3)
        self.windowFunction = QtWidgets.QComboBox(PSDOptionsWidget)
        self.windowFunction.setObjectName("windowFunction")
        self.horizontalLayout_2.addWidget(self.windowFunction)
        self.formLayout.setLayout(6, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout_2)
        self.detrendMethodLabel = QtWidgets.QLabel(PSDOptionsWidget)
        self.detrendMethodLabel.setObjectName("detrendMethodLabel")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.detrendMethodLabel)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem4)
        self.detrendMethod = QtWidgets.QComboBox(PSDOptionsWidget)
        self.detrendMethod.setObjectName("detrendMethod")
        self.horizontalLayout.addWidget(self.detrendMethod)
        self.formLayout.setLayout(7, QtWidgets.QFormLayout.FieldRole, self.horizontalLayout)

        self.retranslateUi(PSDOptionsWidget)
        QtCore.QMetaObject.connectSlotsByName(PSDOptionsWidget)

    def retranslateUi(self, PSDOptionsWidget):
        PSDOptionsWidget.setWindowTitle(QtWidgets.QApplication.translate("PSDOptionsWidget", "PSD Options", None, -1))
        self.fftPointsLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "FFT Window Size", None, -1))
        self.durationLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Window Duration", None, -1))
        self.duration.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "600 s", None, -1))
        self.resolutionLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Frequency Resolution", None, -1))
        self.resolution.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "0.1 Hz", None, -1))
        self.overlapPercentLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Window Overlap Percent", None, -1))
        self.overlapPercent.setSuffix(QtWidgets.QApplication.translate("PSDOptionsWidget", " %", None, -1))
        self.overlapPointsLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Window Overlap Points", None, -1))
        self.windowCountLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Number of Windows", None, -1))
        self.windowCount.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "1", None, -1))
        self.windowFunctionLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Window Function", None, -1))
        self.detrendMethodLabel.setText(QtWidgets.QApplication.translate("PSDOptionsWidget", "Detrend Method", None, -1))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'about.ui',
# licensing of 'about.ui' applies.
#
# Created: Mon May 13 12:31:31 2019
#      by: pyside2-uic  running on PySide2 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(505, 506)
        self.verticalLayout = QtWidgets.QVBoxLayout(AboutDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout.setVerticalSpacing(20)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(AboutDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.softwareLabel = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.softwareLabel.sizePolicy().hasHeightForWidth())
        self.softwareLabel.setSizePolicy(sizePolicy)
        self.softwareLabel.setObjectName("softwareLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.softwareLabel)
        self.label_4 = QtWidgets.QLabel(AboutDialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.label_9 = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_9.sizePolicy().hasHeightForWidth())
        self.label_9.setSizePolicy(sizePolicy)
        self.label_9.setWordWrap(True)
        self.label_9.setObjectName("label_9")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.label_9)
        self.label_5 = QtWidgets.QLabel(AboutDialog)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.label_10 = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_10.sizePolicy().hasHeightForWidth())
        self.label_10.setSizePolicy(sizePolicy)
        self.label_10.setWordWrap(True)
        self.label_10.setObjectName("label_10")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.label_10)
        self.label_6 = QtWidgets.QLabel(AboutDialog)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.label_11 = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_11.sizePolicy().hasHeightForWidth())
        self.label_11.setSizePolicy(sizePolicy)
        self.label_11.setWordWrap(True)
        self.label_11.setObjectName("label_11")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.label_11)
        self.label_7 = QtWidgets.QLabel(AboutDialog)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.label_12 = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_12.sizePolicy().hasHeightForWidth())
        self.label_12.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setWeight(75)
        font.setBold(True)
        self.label_12.setFont(font)
        self.label_12.setWordWrap(True)
        self.label_12.setObjectName("label_12")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.label_12)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(AboutDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AboutDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("accepted()"), AboutDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL("rejected()"), AboutDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(QtWidgets.QApplication.translate("AboutDialog", "About", None, -1))
        self.label.setText(QtWidgets.QApplication.translate("AboutDialog", "Software:", None, -1))
        self.softwareLabel.setText(QtWidgets.QApplication.translate("AboutDialog", "Mondo\n"
"Version: {}\n"
"Build Date: {}", None, -1))
        self.label_4.setText(QtWidgets.QApplication.translate("AboutDialog", "Support:", None, -1))
        self.label_9.setText(QtWidgets.QApplication.translate("AboutDialog", "This software is provided by EPRI \"AS IS\" and without customer support beyond such embodiments within the distribution of this software that may or may not provide such support.", None, -1))
        self.label_5.setText(QtWidgets.QApplication.translate("AboutDialog", "Copyright:", None, -1))
        self.label_10.setText(QtWidgets.QApplication.translate("AboutDialog", "Copyright (c) 2016 Electric Power Research Institute, Inc.\n"
"Copyright (c) 2019 Suprock Technologies, LLC\n"
"All rights reserved.\n"
"\n"
"Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.", None, -1))
        self.label_6.setText(QtWidgets.QApplication.translate("AboutDialog", "Developed by:", None, -1))
        self.label_11.setText(QtWidgets.QApplication.translate("AboutDialog", "Suprock Technologies\n"
"45 Scott Hill Rd\n"
"Warren, NH 03279", None, -1))
        self.label_7.setText(QtWidgets.QApplication.translate("AboutDialog", "Disclaimer:", None, -1))
        self.label_12.setText(QtWidgets.QApplication.translate("AboutDialog", "THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EPRI BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.", None, -1))




# Copyright (c) 2016 Electric Power Research Institute, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the EPRI nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import math

import numpy
from PySide2 import QtWidgets
import matplotlib.mlab as mlab

from . import util
from .. import export_csv
from .. import export_script

logger = logging.getLogger(__name__)


def do_psd(sequence, subchannel_index, NFFT, Fs, detrend, window, noverlap):
    # JSN NOTE: much of this function was taken from matplotlib.mlab's
    # internal _spectral_helper(). It has been streamlined by removing
    # unneeded features, and changed to support chunks of data

    freqs = numpy.fft.rfftfreq(NFFT, 1 / Fs)

    windows = 0
    sums = numpy.zeros(freqs.shape)

    for _time, data, _start, _end in sequence:
        x = numpy.asarray(data[:, subchannel_index])

        if len(x) < NFFT:
            continue

        result = mlab.stride_windows(x, NFFT, noverlap, axis=0)
        result = mlab.detrend(result, detrend, axis=0)
        result, windowVals = mlab.apply_window(result, window, axis=0,
                                               return_window=True)
        result = numpy.fft.rfft(result, n=NFFT, axis=0)

        result = (numpy.conjugate(result) * result).real

        # Scale the spectrum by the norm of the window to compensate for
        # windowing loss; see Bendat & Piersol Sec 11.5.2.
        result /= (numpy.abs(windowVals) ** 2).sum()

        windows += result.shape[1]
        result = result.sum(axis=1)
        sums += result

        del result

    sums[1:] *= 2.0  # because it's one sided

    sums /= Fs  # get the spectral density

    sums /= windows

    return sums, freqs


def psd_analysis(parent=None):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    channel_index = util.choose_channel(header, parent=parent)
    if channel_index is None:
        return  # error message displayed, or user cancelled

    batch_info = util.decode_batch(file_infos, header, [channel_index],
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    subchannel_index = util.choose_subchannel(batch_info[channel_index],
                                              allow_all=False)
    if subchannel_index is None:
        return  # user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, _unit_info = util.sequence_data(
        batch_info, chunk=True, raw_units=True, parent=parent)

    sequence = sequence_info[channel_index]

    sampling_rate = None
    for time, _data, _start, _end in sequence:
        if len(time) > 2:
            sampling_rate = 1.0 / (time[1] - time[0])
            break

    if sampling_rate is None:
        message = "Could not determine sampling rate!"
        logger.error(message)
        QtWidgets.QMessageBox.critical(parent, "Error", message)
        return

    psd_options = util.get_psd_options([sequence], sampling_rate,
                                       parent=parent)
    if psd_options is None:
        return  # user cancelled

    pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
    pxx_dB = 10 * numpy.log10(pxx)

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    ax.plot(freqs, pxx_dB)
    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        logi = 0.1
    step = 10 * logi
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    export_csv.add_export_csv_action(fig, labels, [(freqs, pxx_dB, None)])

    fig.show()


def single_channel_psd_analysis(parent=None):
    channel_index = None  # ask for the channel on the first iteration
    subchannel_index = None  # ask for the subchannel on the first iteration

    sampling_rate = None
    sequences = []
    names = []
    files = set()

    while True:
        # load a batch of files
        ret = util.load_batch(parent=parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # warn the user if any files have been loaded before
        for filename, _dt in file_infos:
            if filename in files:
                message = ("File {} already loaded in another batch!\n"
                           "Proceed with caution!").format(filename)
                logger.warning(message)
                QtWidgets.QMessageBox.warning(parent, "Warning", message)
            files.add(filename)

        # if it's the first time through, choose a channel
        if channel_index is None:
            channel_index = util.choose_channel(header, parent=parent)
            if channel_index is None:
                return  # error message displayed, or user cancelled

        # make sure this batch has a channel of the correct index
        if channel_index >= len(header['channels']):
            message = "Channels inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        batch_info = util.decode_batch(file_infos, header, [channel_index],
                                       parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        if subchannel_index is None:
            subchannel_index = util.choose_subchannel(
                batch_info[channel_index], allow_all=False)
            if subchannel_index is None:
                return  # user cancelled

        subchannels = batch_info[channel_index].channel_decoder.subchannels
        if subchannel_index >= subchannels:
            message = "Channels inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        batch_info = util.get_datetime_subset(batch_info, parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret = util.warn_about_lost_packets(batch_info, parent=parent)
        if not ret:
            return  # user cancelled

        sequence_info, _unit_info = util.sequence_data(
            batch_info, chunk=True, raw_units=True, parent=parent)

        sequence = sequence_info[channel_index]

        sequences.append(sequence)

        batch_sampling_rate = None
        for time, _data, _start, _end in sequence:
            if len(time) > 2:
                batch_sampling_rate = 1.0 / (time[1] - time[0])
                break

        if batch_sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        if sampling_rate is None:
            sampling_rate = batch_sampling_rate
        elif sampling_rate != batch_sampling_rate:
            message = "Sampling rates inconsistent between batches!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
            default=QtWidgets.QMessageBox.No)
        if ret != QtWidgets.QMessageBox.Yes:
            break

    psd_options = util.get_psd_options(sequences, sampling_rate, parent=parent)
    if psd_options is None:
        return  # user cancelled

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []

    for i, sequence in enumerate(sequences):
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=names[i])
        export_data.append((freqs, pxx_dB, names[i]))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        logi = 0.1
    step = 10 * logi
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    export_csv.add_export_csv_action(fig, labels, export_data)

    fig.show()


def single_slice_psd_analysis(parent=None):
    ret = util.load_batch(parent=parent)
    if ret is None:
        return  # error message displayed, or user cancelled
    file_infos, header = ret

    subchannels_list = util.choose_subchannels(header, parent=parent)
    if subchannels_list is None:
        return  # error message displayed, or user cancelled

    channel_indexes = {t[0] for t in subchannels_list}

    batch_info = util.decode_batch(file_infos, header, channel_indexes,
                                   parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    batch_info = util.get_datetime_subset(batch_info, parent=parent)
    if batch_info is None:
        return  # error message displayed, or user cancelled

    ret = util.warn_about_lost_packets(batch_info, parent=parent)
    if not ret:
        return  # user cancelled

    sequence_info, _unit_info = util.sequence_data(
        batch_info, chunk=True, raw_units=True, parent=parent)

    sampling_rates = {}  # key: channel, value: sampling rate
    for channel_index in channel_indexes:
        sampling_rate = None
        for time, _data, _start, _end in sequence_info[channel_index]:
            if len(time) > 2:
                sampling_rate = 1.0 / (time[1] - time[0])
                break

        if sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        sampling_rates[channel_index] = sampling_rate

    # create a list of similar channels, which can share psd options
    channel_signatures = {}
    for channel_index in channel_indexes:
        sequence = sequence_info[channel_index]
        sequence_signature = tuple(len(x[0]) for x in sequence)
        channel_signatures[channel_index] = (sampling_rates[channel_index],
                                             sequence_signature)

    remaining_channels = set(channel_indexes)
    sections = []
    channel_groups = []
    for channel_index in sorted(channel_indexes):
        if channel_index in remaining_channels:
            remaining_channels.discard(channel_index)
        else:
            continue

        sig = channel_signatures[channel_index]
        similar_channels = [channel_index]
        for other_channel_index in sorted(remaining_channels):
            if sig == channel_signatures[other_channel_index]:
                similar_channels.append(other_channel_index)
                remaining_channels.discard(other_channel_index)

        group_name = ", ".join((batch_info[i].channel_decoder.channel_name
                                for i in similar_channels))

        # the sequence and rate of all the similar channels are the same
        sequence = sequence_info[similar_channels[0]]
        sampling_rate = sampling_rates[similar_channels[0]]

        sections.append((group_name, [sequence], sampling_rate))
        channel_groups.append(similar_channels)

    all_psd_options = util.get_multiple_psd_options(sections, parent=parent)
    if all_psd_options is None:
        return  # user cancelled

    # unpack psd options, per channel
    channel_psd_options = {}
    for i, psd_options in enumerate(all_psd_options):
        for channel_index in channel_groups[i]:
            channel_psd_options[channel_index] = psd_options

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []

    for channel_index, subchannel_index in subchannels_list:
        channel_decoder = batch_info[channel_index].channel_decoder
        if len(channel_decoder.subchannel_names) > 1:
            name = channel_decoder.subchannel_names[subchannel_index]
        else:
            name = channel_decoder.channel_name

        sequence = sequence_info[channel_index]
        psd_options = channel_psd_options[channel_index]
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=name)
        export_data.append((freqs, pxx_dB, name))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        logi = 0.1
    step = 10 * logi
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    export_csv.add_export_csv_action(fig, labels, export_data)

    fig.show()


def overlaid_psd_analysis(parent=None):
    sequences = []
    names = []
    sampling_rates = []

    while True:
        # load a batch of files
        ret = util.load_batch(parent=parent)
        if ret is None:
            return  # error message displayed, or user cancelled
        file_infos, header = ret

        # if it's the first time through, choose a channel
        channel_index = util.choose_channel(header, parent=parent)
        if channel_index is None:
            return  # error message displayed, or user cancelled

        batch_info = util.decode_batch(file_infos, header, [channel_index],
                                       parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        subchannel_index = util.choose_subchannel(
            batch_info[channel_index], allow_all=False)
        if subchannel_index is None:
            return  # user cancelled

        batch_info = util.get_datetime_subset(batch_info, parent=parent)
        if batch_info is None:
            return  # error message displayed, or user cancelled

        ret = util.warn_about_lost_packets(batch_info, parent=parent)
        if not ret:
            return  # user cancelled

        sequence_info, _unit_info = util.sequence_data(
            batch_info, chunk=True, raw_units=True, parent=parent)

        sequence = sequence_info[channel_index]

        sequences.append(sequence)

        sampling_rate = None
        for time, _data, _start, _end in sequence:
            if len(time) > 2:
                sampling_rate = 1.0 / (time[1] - time[0])
                break

        if sampling_rate is None:
            message = "Could not determine sampling rate!"
            logger.error(message)
            QtWidgets.QMessageBox.critical(parent, "Error", message)
            return

        sampling_rates.append(sampling_rate)

        # ask the user for a name for this set
        msg = "Display name for this batch (#{}):".format(len(sequences))
        name, ok = QtWidgets.QInputDialog.getText(parent, "Batch Name", msg)
        if not ok:
            return  # user cancelled

        names.append(name.strip())

        # ask the user if they want to load more
        msg = "Load more batches? Loaded {} so far".format(len(sequences))
        ret = QtWidgets.QMessageBox.question(
            parent, "More Batches?", msg,
            buttons=(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No),
            default=QtWidgets.QMessageBox.No)
        if ret != QtWidgets.QMessageBox.Yes:
            break

    sections = []
    for i, sequence in enumerate(sequences):
        sections.append((names[i], [sequence], sampling_rates[i]))

    all_psd_options = util.get_multiple_psd_options(sections, parent=parent)
    if all_psd_options is None:
        return  # user cancelled

    # wrapper around matplotlib.pyplot.subplots()
    fig, ax = export_script.subplots_wrapper()

    export_data = []

    for i, sequence in enumerate(sequences):
        psd_options = all_psd_options[i]
        pxx, freqs = do_psd(sequence, subchannel_index, **psd_options)
        pxx_dB = 10 * numpy.log10(pxx)
        ax.plot(freqs, pxx_dB, label=names[i])
        export_data.append((freqs, pxx_dB, names[i]))

    ax.legend()

    ax.grid(True)

    vmin, vmax = ax.viewLim.intervaly
    logi = int(numpy.log10(vmax - vmin))
    if logi == 0:
        logi = 0.1
    step = 10 * logi
    ticks = numpy.arange(math.floor(vmin), math.ceil(vmax) + 1, step)
    ax.set_yticks(ticks)

    labels = ("Frequency (Hz)", "Power Spectral Density (dB)")

    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    ax.get_yaxis().get_major_formatter().set_useOffset(False)

    export_csv.add_export_csv_action(fig, labels, export_data)

    fig.show()
